from pathlib import Path
from typing import AsyncContextManager, Dict, List, Literal, Optional, Pattern, Union

from playwright._impl._api_structures import (
    Geolocation,
    HttpCredentials,
    ProxySettings,
    StorageState,
    ViewportSize,
)
from playwright.async_api import Browser, BrowserContext, Page
from typing_extensions import TypedDict, Unpack


class Parameters(TypedDict, total=False):
    viewport: Optional[ViewportSize]
    screen: Optional[ViewportSize]
    no_viewport: Optional[bool]
    ignore_https_errors: Optional[bool]
    java_script_enabled: Optional[bool]
    bypass_csp: Optional[bool]
    user_agent: Optional[str]
    locale: Optional[str]
    timezone_id: Optional[str]
    geolocation: Optional[Geolocation]
    permissions: Optional[List[str]]
    extra_http_headers: Optional[Dict[str, str]]
    offline: Optional[bool]
    http_credentials: Optional[HttpCredentials]
    device_scale_factor: Optional[float]
    is_mobile: Optional[bool]
    has_touch: Optional[bool]
    color_scheme: Optional[Literal["dark", "light", "no-preference"]]
    forced_colors: Optional[Literal["active", "none"]]
    reduced_motion: Optional[Literal["no-preference", "reduce"]]
    accept_downloads: Optional[bool]
    default_browser_type: Optional[str]
    proxy: Optional[ProxySettings]
    record_har_path: Optional[Union[str, Path]]
    record_har_omit_content: Optional[bool]
    record_video_dir: Optional[Union[str, Path]]
    record_video_size: Optional[ViewportSize]
    storage_state: Optional[Union[StorageState, str, Path]]
    base_url: Optional[str]
    strict_selectors: Optional[bool]
    service_workers: Optional[Literal["allow", "block"]]
    record_har_url_filter: Optional[Union[str, Pattern[str]]]
    record_har_mode: Optional[Literal["full", "minimal"]]
    record_har_content: Optional[Literal["attach", "embed", "omit"]]


class RegularPage(AsyncContextManager[Page]):
    browser: Browser
    page: Page

    def __init__(self, browser: Browser, **kwargs: Unpack[Parameters]):
        self.browser = browser
        self.kwargs = kwargs

    async def __aenter__(self) -> Page:
        self.page = await self.browser.new_page(**self.kwargs)
        return self.page

    async def __aexit__(self, *args) -> None:
        await self.page.close()
        return None


class ContextualPage(AsyncContextManager[Page]):
    browser: Browser
    context: BrowserContext
    page: Page

    def __init__(self, browser: Browser, **kwargs: Unpack[Parameters]):
        self.browser = browser
        self.kwargs = kwargs

    async def __aenter__(self) -> Page:
        self.context = await self.browser.new_context(**self.kwargs)
        self.page = await self.context.new_page()
        return self.page

    async def __aexit__(self, *args) -> None:
        await self.page.close()
        await self.context.close()
        return None



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from launart import Service
from playwright._impl._api_structures import ProxySettings
from playwright.async_api import Browser, Playwright, async_playwright
from typing_extensions import ParamSpec

from .installer import install_playwright
from .interface import PlaywrightBrowser, PlaywrightBrowserImpl

P = ParamSpec("P")


class PlaywrightService(Service):
    """用于 launart 的浏览器服务

    Args:
        browser_type(Literal["chromium", "firefox", "webkit"]): 你要使用的浏览器，默认为 Chromium
        **kwargs: 参见 <https://playwright.dev/python/docs/api/class-browsertype#browser-type-launch>
    """

    id = "web.render/playwright"
    supported_interface_types = {PlaywrightBrowser}
    playwright: Playwright
    browser: Browser
    browser_type: str
    launch_config: Dict[str, Any]
    playwright_download_host: Optional[str]

    if TYPE_CHECKING:

        def __init__(
            self,
            browser_type: Literal["chromium", "firefox", "webkit"] = "chromium",
            playwright_download_host: Optional[str] = None,
            *,
            executable_path: Optional[Union[str, Path]] = None,
            channel: Optional[str] = None,
            args: Optional[List[str]] = None,
            ignore_default_args: Optional[Union[bool, List[str]]] = None,
            handle_sigint: Optional[bool] = None,
            handle_sigterm: Optional[bool] = None,
            handle_sighup: Optional[bool] = None,
            timeout: Optional[float] = None,
            env: Optional[Dict[str, Union[str, float, bool]]] = None,
            headless: Optional[bool] = None,
            devtools: Optional[bool] = None,
            proxy: Optional[ProxySettings] = None,
            downloads_path: Optional[Union[str, Path]] = None,
            slow_mo: Optional[float] = None,
            traces_dir: Optional[Union[str, Path]] = None,
            chromium_sandbox: Optional[bool] = None,
            firefox_user_prefs: Optional[Dict[str, Union[str, float, bool]]] = None,
        ):
            ...

    else:

        def __init__(
            self,
            browser_type: Literal["chromium", "firefox", "webkit"] = "chromium",
            playwright_download_host: Optional[str] = None,
            **kwargs,
        ) -> None:
            self.browser_type: Literal["chromium", "firefox", "webkit"] = browser_type
            self.launch_config = kwargs
            self.playwright_download_host = playwright_download_host
            super().__init__()

    def get_interface(self, _):
        return PlaywrightBrowserImpl(self, self.browser)

    @property
    def required(self):
        return set()

    @property
    def stages(self):
        return {"preparing", "cleanup"}

    async def launch(self, _):
        await install_playwright(self.playwright_download_host, self.browser_type)
        playwright_mgr = async_playwright()

        async with self.stage("preparing"):
            self.playwright = await playwright_mgr.__aenter__()
            browser_type = {
                "chromium": self.playwright.chromium,
                "firefox": self.playwright.firefox,
                "webkit": self.playwright.webkit,
            }[self.browser_type]
            self.browser = await browser_type.launch(**self.launch_config)

        async with self.stage("cleanup"):
            await playwright_mgr.__aexit__()



import asyncio
import re
from typing import Optional

from playwright._impl._driver import compute_driver_executable, get_driver_env

from .utils import Progress, log

download_complete = re.compile("(?P<file>.*) downloaded to (?P<path>.*)")
percent_pat = re.compile("(\\d+)%")


async def install_playwright(download_host: Optional[str] = None, browser_type: str = "chromium"):
    command = [str(compute_driver_executable()), "install", browser_type]
    env = get_driver_env()
    if download_host:
        env["PLAYWRIGHT_DOWNLOAD_HOST"] = download_host
    shell = await asyncio.create_subprocess_exec(*command, stdout=asyncio.subprocess.PIPE, env=env)
    assert shell.stdout

    progress: Optional[Progress] = None

    while line := (await shell.stdout.readline()).decode("UTF-8"):
        if "Downloading" in line:
            progress = Progress(line[12:-1])
        if percent := percent_pat.findall(line):
            progress_target = float(percent[0])
            if progress:
                progress.update(target=progress_target)
        elif p := download_complete.match(line):
            p = p.groupdict()
            log("success", f"Downloaded [cyan]{p['file']}[/] to [magenta]{p['path']}[/]")
        elif line == "Failed to install browsers\n":
            message = await shell.stdout.read()
            log("error", "Download Failed:\n" + message.decode("UTF-8"))

    if shell.returncode:
        log("error", "failed to download Playwright.")
        log("error", "Please see: [magenta]https://playwright.dev/python/docs/intro[/]")
        log(
            "error",
            "Run [magenta]poetry run playwright install[/] or [magenta]pdm run playwright install[/] to install Playwright manually.",
        )
    else:
        log("success", f"Playwright for {browser_type} is ready.")
